MODULE MEuler

USE MUserFunc, ONLY: f => f, IsInDomain_f => IsInDomain_f

IMPLICIT NONE

! ERROR list
INTEGER, PARAMETER :: MEULER_ERROR_STANDARD = -1
INTEGER, PARAMETER :: MEULER_ERROR_GLOBS = -2
INTEGER, PARAMETER :: MEULER_ERROR_OUTOFDOMAIN = -3
INTEGER, PARAMETER :: MEULER_ERROR_OVERFLOW = -4

! buffer length for string operations
INTEGER, PARAMETER :: MEULER_MSG_LENGTH = 256

! parameter for f
DOUBLE PRECISION :: b

! increment
DOUBLE PRECISION :: h

! initial value
DOUBLE PRECISION :: t0
DOUBLE PRECISION :: eta

! arrays for the polygonal methods
INTEGER :: nDim
DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: tArray
DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: yArray

! swap globs
COMMON /MEULER_GLOBS/ h, t0, b, eta

CONTAINS

! This function checks the sum a + b on possible overflow.
! It returns .TRUE. on overflow, otherwise .FALSE.
LOGICAL FUNCTION CheckSumOverflow(a, b)
   DOUBLE PRECISION, INTENT(IN) :: a, b

   DOUBLE PRECISION :: alpha, beta, x
   
   ! fast exits
   IF ((a .EQ. DBLE(0)) .OR. (b .EQ. DBLE(0))) THEN
      CheckSumOverflow = .FALSE.
      RETURN
   END IF
   If ((a .GT. DBLE(0)) .AND. (b .LT. DBLE(0))) THEN
      CheckSumOverflow = .FALSE.
      RETURN
   END IF
   If ((a .LT. DBLE(0)) .AND. (b .GT. DBLE(0))) THEN
      CheckSumOverflow = .FALSE.
      RETURN
   END IF
   ! here is (DABS(a + b) .EQ. (DABS(a) + DABS(b)))

   alpha = DABS(a)
   beta = DABS(b)

   ! no overflow in the following x
   x = HUGE(alpha) - alpha
   ! here is ((alpha + x) .EQ. HUGE(alpha))
   IF (beta .GT. x) THEN
      ! here is ((alpha + beta) .GT. (alpha + x))
      ! i.e. ((alpha + beta) .GT. HUGE(alpha))
      ! Overflow!
      CheckSumOverflow = .TRUE.
      RETURN
   END IF

   CheckSumOverflow = .FALSE.
   RETURN
END FUNCTION CheckSumOverflow

! This function checks the product a * b on possible overflow.
! It returns .TRUE. on overflow, otherwise .FALSE.
LOGICAL FUNCTION CheckProductOverflow(a, b)
   DOUBLE PRECISION, INTENT(IN) :: a, b

   DOUBLE PRECISION :: alpha, beta, x

   alpha = DABS(a)
   beta = DABS(b)

   ! fast exit
   IF ((alpha .LE. DBLE(1)) .OR. (beta .LE. DBLE(1))) THEN
      CheckProductOverflow = .FALSE.
      RETURN
   END IF
   ! here is ((alpha .GT. DBLE(1)) .AND. (beta .GT. DBLE(1)))

   ! no overflow in the following x
   x = HUGE(alpha) / alpha
   ! here is ((alpha * x) .EQ. HUGE(alpha))
   IF (beta .GT. x) THEN
      ! here is ((alpha * beta) .GT. (alpha * x))
      ! i.e. ((alpha * beta) .GT. HUGE(alpha))
      CheckProductOverflow = .TRUE.
      RETURN
   END IF

   CheckProductOverflow = .FALSE.
   RETURN
END FUNCTION CheckProductOverflow

! This function checks the global variables!
! It returns .FALSE., if anything is wrong, otherwise .TRUE.
LOGICAL FUNCTION CheckGlobs()
   ! check increment
   IF (h .EQ. DBLE(0)) THEN
      PRINT*, "ERROR: Increment h = 0.0"
      CheckGlobs = .FALSE.
      RETURN
   END IF

   ! check t0, b
   IF (h .GT. DBLE(0)) THEN
      IF (b .LE. t0) THEN
         PRINT*,"ERROR: h > 0 and b <= t0!"
         CheckGlobs = .FALSE.
         RETURN
      END IF
   ELSE
      ! here is (h .LT. DBLE(0))
      IF (b .GE. t0) THEN
         PRINT*,"ERROR: h < 0 and b >= t0!"
         CheckGlobs = .FALSE.
         RETURN
      END IF
   END IF

   ! check initial value
   IF (.NOT. IsInDomain_f(eta, t0)) THEN
      PRINT*, "ERROR: Initial value out of range: y(", t0, ") = ", eta
      CheckGlobs = .FALSE.
      RETURN
   END IF

   CheckGlobs = .TRUE.
   RETURN
END FUNCTION CheckGlobs

! This subroutine prints the data of a polygonal method.
SUBROUTINE PrintData(msg)
   CHARACTER( len = MEULER_MSG_LENGTH ), INTENT(IN) :: msg

   INTEGER :: i

   IF ((LBOUND(tArray, 1) .NE. LBOUND(yArray, 1)) .OR. (UBOUND(tArray, 1) .NE. UBOUND(yArray, 1))) THEN
      PRINT*,"Error in allocation!"
      RETURN
   END IF

   IF (nDim .LT. 0) THEN
      PRINT*,"No DATA!"
      RETURN
   END IF
   IF (nDim .GT. UBound(tArray, 1)) THEN
      PRINT*,"Error in alloaction!"
      RETURN
   END IF
   
   IF (nDim .EQ. HUGE(nDim)) THEN
      PRINT*,"Overflow!"
      RETURN
   END IF

   PRINT*, msg, (nDim + 1), " items"

   ! print globs
   IF (h .GT. DBLE(0)) THEN
      ! here is (t0 .LT. b)
      PRINT*,"Domain(f) = R x [", t0 , "; ", b, "]"
   ELSE
      ! here is (b .LT. t0)
      PRINT*,"Domain(f) = R x [", b , "; ", t0, "]"
   END IF

   PRINT*,"Increment h = ", h
   PRINT*,"Initial value: y(", t0, ") = ", eta

   IF (h .GT. DBLE(0)) THEN
      i = LBOUND(tArray, 1)
      DO WHILE (i .LE. nDim)
         PRINT*,"Step i = ", i, ": (y_i; t_i) = (", yArray(i), "; ", tArray(i), ")"

         ! check for overflow
         IF (i .EQ. HUGE(i)) THEN
            PRINT*,"Overflow!"
            RETURN
         ENDIF
         ! next item
         i = i + 1
      END DO
   ELSE
      ! here is (h .LT. DBLE(0))
      i = nDim
      DO WHILE (i .GE. LBOUND(tArray, 1))
         PRINT*,"Step i = ", i, ": (y_i; t_i) = (", yArray(i), "; ", tArray(i), ")"

         ! next item
         i = i - 1
      END DO
   END IF

   RETURN
END SUBROUTINE PrintData

! THis function is an implementation of Euler's Polygonal Method.
! It returns a value < 0, if anything is wrong.
! It returns MEULER_ERROR_GLOBS, if the global variables are not OK.
! It returns MEULER_ERROR_OUTOFDOMAIN, if a pair (y, t) is not in the domain of f.
! It returns MEULER_ERROR_OVERFLOW, if an overflow occurs.
INTEGER FUNCTION DoRecursionOld()
   DOUBLE PRECISION :: y, t
   DOUBLE PRECISION :: alpha, beta, gamma
   INTEGER :: i
   INTEGER :: nMax
   LOGICAL :: OutOfDomain

   ! check globs
   IF (.NOT. CheckGlobs()) THEN
      DoRecursionOld = MEULER_ERROR_GLOBS
      RETURN
   END IF

   ! compute maximum number of steps
   alpha = (b - t0) / h
   nMax = FLOOR(alpha)
   ! check nMax
   IF (nMax .LT. 0) THEN
      DoRecursionOld = MEULER_ERROR_STANDARD
      RETURN
   END IF

   ! allocate memory
   ALLOCATE(tArray(0:nMax))
   ALLOCATE(yArray(0:nMax))
   nDim = 0

   ! clear arrays
   i = 0
   DO WHILE (i <= nMax)

      tArray(i) = DBLE(0)
      yArray(i) = DBLE(0)

      ! check for overflow
      IF (i .EQ. HUGE(i)) THEN
         DoRecursionOld = MEULER_ERROR_OVERFLOW
         RETURN
      END IF

      ! next item
      i = i + 1
   END DO

   ! beginning of recursion
   ! check initial values
   IF (.NOT. IsInDomain_f(eta, t0)) THEN
      DoRecursionOld = MEULER_ERROR_OUTOFDOMAIN
      RETURN
   END IF
   i = 0
   tArray(i) = t0
   yArray(i) = eta

   DO
      ! check boundaries
      IF ((i + 1) .GT. nMax) THEN
         nDim = nMax
         DoRecursionOld = nDim
         RETURN
      END IF

      ! check (tArray(i) + h) for overflow
      IF (CheckSumOverflow(tArray(i), h)) THEN
         DoRecursionOld = MEULER_ERROR_OVERFLOW
         RETURN
      END IF

      beta = f(OutOfDomain, yArray(i), tArray(i))
      ! check for OutOfDomain
      if (OutOfDomain) THEN
         DoRecursionOld = MEULER_ERROR_OUTOFDOMAIN
         RETURN
      END IF

      ! check (h * beta) for overflow
      IF (CheckProductOverflow(h, beta)) THEN
         DoRecursionOld = MEULER_ERROR_OVERFLOW
         RETURN
      END IF
      gamma = h * beta

      ! check  (yArray(i) + gamma) for overflow
      IF (CheckSumOverflow(yArray(i), gamma)) THEN
         DoRecursionOld = MEULER_ERROR_OVERFLOW
         RETURN
      END IF

      ! do the recursion
      t = tArray(i) + h
      y = yArray(i) + gamma

      ! check domain for break
      IF (.NOT. IsInDomain_f(y, t)) THEN
         nDim = i
         DoRecursionOld = nDim
         RETURN
      END IF

      ! store (y, t) in arrays
      tArray(i + 1) = t
      yArray(i + 1) = y

      ! next item (check for overflow)
      IF (i .EQ. HUGE(i)) THEN
         DoRecursionOld = MEULER_ERROR_OVERFLOW
         RETURN
      ENDIF
      i = i + 1
   END DO
   
   ! this caonnot happen
   DoRecursionOld = MEULER_ERROR_STANDARD
   RETURN
END FUNCTION DoRecursionOld

! THis function is an implementation of the new polygonal method.
! It returns a value < 0, if anything is wrong.
! It returns MEULER_ERROR_GLOBS, if the global variables are not OK.
! It returns MEULER_ERROR_OUTOFDOMAIN, if a pair (y, t) is not in the domain of f.
! It returns MEULER_ERROR_OVERFLOW, if an overflow occurs.
INTEGER FUNCTION DoRecursionNew(nMax)
   INTEGER, INTENT(IN) :: nMax

   DOUBLE PRECISION :: y, t
   DOUBLE PRECISION :: alpha, beta, gamma, delta
   INTEGER :: i
   LOGICAL :: OutOfDomain

   ! check globs
   IF (.NOT. CheckGlobs()) THEN
      DoRecursionNew = MEULER_ERROR_GLOBS
      RETURN
   END IF

   ! exit with MEULER_ERROR_STANDARD on error
   IF (nMax .LT. 0) THEN
      DoRecursionNew = MEULER_ERROR_STANDARD
      RETURN
   END IF

   ! allocate memory
   ALLOCATE(tArray(0:nMax))
   ALLOCATE(yArray(0:nMax))
   nDim = 0

   ! clear arrays
   i = 0
   DO WHILE (i <= nMax)

      tArray(i) = DBLE(0)
      yArray(i) = DBLE(0)

      ! check for overflow
      IF (i .EQ. HUGE(i)) THEN
         DoRecursionNew = MEULER_ERROR_OVERFLOW
         RETURN
      END IF

      ! next item
      i = i + 1
   END DO

   ! beginning of recursion
   ! check initial values
   IF (.NOT. IsInDomain_f(eta, t0)) THEN
      DoRecursionNew = MEULER_ERROR_OUTOFDOMAIN
      RETURN
   END IF
   i = 0
   tArray(i) = t0
   yArray(i) = eta

   DO
      ! check boundaries
      IF ((i + 1) .GT. nMax) THEN
         nDim = nMax
         DoRecursionNew = nDim
         RETURN
      END IF

      alpha = f(OutOfDomain, yArray(i), tArray(i))
      ! check for OutOfDomain
      if (OutOfDomain) THEN
         DoRecursionNew = MEULER_ERROR_OUTOFDOMAIN
         RETURN
      END IF

      ! check alpha * alpha for overflow
      IF (CheckProductOverflow(alpha, alpha)) THEN
         DoRecursionNew = MEULER_ERROR_OVERFLOW
         RETURN
      END IF
      beta = alpha * alpha
      
      ! check DBLE(1) + beta for overflow
      IF (CheckSumOverflow(DBLE(1), beta)) THEN
         DoRecursionNew = MEULER_ERROR_OVERFLOW
         RETURN
      END IF
      beta = DSQRT(DBLE(1) + beta)
      ! here is (beta .GE. DBLE(1))
      ! so (DBLE(1) / beta) is a defined value
      
      ! check (h / beta) for overflow
      IF (CheckProductOverflow(h, DBLE(1) / beta)) THEN
         DoRecursionNew = MEULER_ERROR_OVERFLOW
         RETURN
      END IF
      gamma = h / beta
      
      ! check gamma * alpha for overflow
      IF (CheckProductOverflow(gamma, alpha)) THEN
         DoREcursionNew = MEULER_ERROR_OVERFLOW
         RETURN
      END IF
      delta = gamma * alpha

      ! check (tArray(i) + gamma) for overflow
      IF (CheckSumOverflow(tArray(i), gamma)) THEN
         DoRecursionNew = MEULER_ERROR_OVERFLOW
         RETURN
      END IF

      ! check (yArray(i) + delta) for overflow
      IF (CheckSumOverflow(yArray(i), delta)) THEN
         DoRecursionNew = MEULER_ERROR_OVERFLOW
         RETURN
      END IF

      ! do the recursion
      t = tArray(i) + gamma
      y = yArray(i) + delta

      ! check domain for break
      IF (.NOT. IsInDomain_f(y, t)) THEN
         nDim = i
         DoRecursionNew = nDim
         RETURN
      END IF

      ! store (y, t) in arrays
      tArray(i + 1) = t
      yArray(i + 1) = y

      ! next item (check for overflow)
      IF (i .EQ. HUGE(i)) THEN
         DoRecursionNew = MEULER_ERROR_OVERFLOW
         RETURN
      ENDIF
      i = i + 1
   END DO
   
   ! this caonnot happen
   DoRecursionNew = MEULER_ERROR_STANDARD
   RETURN
END FUNCTION DoRecursionNew

! free memory
SUBROUTINE ClearGlobs()
   DEALLOCATE(yArray)
   DEALLOCATE(tArray)

   RETURN
END SUBROUTINE ClearGlobs

END MODULE MEuler
